"""Various implementations of the Multiband Lomb-Scargle Periodogram"""

from .main import available_methods, lombscargle_multiband
from .mbfast_impl import lombscargle_mbfast
from .mbflex_impl import lombscargle_mbflex
